import inspect
from functools import wraps

import scipy
import numpy as np
import pandas as pd

import lmfit
from lmfit import Parameter, Parameters

from epistasis.stats import gmean, pearson
from epistasis.models.utils import arghandler
from epistasis.models.linear.ordinary import EpistasisLinearRegression
from epistasis.models.nonlinear.ordinary import EpistasisNonlinearRegression

from .minimizer import FunctionMinimizer

# -------------------- Power Transform Function -----------------------

def power_transform(x, lmbda, A, B, data=None):
    """Transform x according to a power transformation.

    Note, this functions calculates the geometric mean of x
    to center the power transform on the data. If you'd like to calculate
    the geometric mean on a different array than x (perhaps some
    other data) pass that ohter array to the data keyword argument.

    .. math::
        y = \\frac{ x^{\\lambda} - 1 }{\\lambda [GM(x)]^{\\lambda - 1}}

    Parameters
    ----------
    x : array-like
        data to transform.
    lmbda : float
        power parameter.
    A : float
        horizontal translation constant.
    B : float
        vertical translation constant.
    data : array-like (default=None)
        data to calculate the geometric mean.
    """
    # Calculate the GMean on the data
    if data is None:
        gm = gmean(x + A)
    else:
        gm = gmean(data + A)

    # Check for zeros
    if lmbda == 0:
        return gm * np.log(x + A)
    else:
        first = (x + A)**lmbda
        out = (first - 1.0) / (lmbda * gm**(lmbda - 1)) + B

    return out

# --------------------- Power transform Minizer object -----------------------

class PowerTransformMinizer(FunctionMinimizer):
    """Minimizer class for power transform.
    """
    def __init__(self, **p0):
        # Construct parameters object
        self.parameters = Parameters()
        for p in ['lmbda', 'A', 'B']:
            # Get starting value of parameter if given.
            val = None
            if p in p0:
                val = p0[p]
            # Add parameter.
            self.parameters.add(name=p, value=val)

        # Set function
        self._function = power_transform

    def function(self, x, lmbda, A, B):
        """Execute the function."""
        return self._function(x, lmbda=lmbda, A=A, B=B, data=self.data)

    def predict(self, x):
        return self._function(x, **self.parameter, data=self.data)

    def fit(self, x, y):
        self.data = x

        # Set the lower bound on B.
        self.parameters['A'].set(min=-min(x))

        # Store residual steps in case fit fails.
        last_residual_set = None

        # Residual function to minimize.
        def residual(params, func, x, y=None, data=None):
            # Fit model
            parvals = list(params.values())
            ymodel = func(x, *parvals, data=data)

            # Store items in case of error.
            nonlocal last_residual_set
            last_residual_set = (params, ymodel)

            return y - ymodel

        # Minimize the above residual function.
        try:
            self.minimizer = lmfit.minimize(residual, self.parameters,
                                            args=[self._function, x],
                                            kws={'y': y, 'data': self.data})
        # If fitting fails, print what happened
        except Exception as e:
            # if e is ValueError
            print("ERROR! Some of the transformed phenotypes are invalid.")
            print("\nParameters:")
            print("----------")
            print(last_residual_set[0].pretty_print())
            print("\nTransformed phenotypes:")
            print("----------------------")
            print(last_residual_set[1])
            raise

        self.parameters = self.minimizer.params


# -------------------- Epistasis Model -----------------------


class EpistasisPowerTransform(EpistasisNonlinearRegression):
    """Use power-transform function, via nonlinear least-squares regression,
    to estimate epistatic coefficients and the nonlinear scale in a nonlinear
    genotype-phenotype map.

    This models has two steps:
        1. Fit an additive, linear regression to approximate the average effect
        of individual mutations.
        2. Fit the nonlinear function to the observed phenotypes vs. the
        additive phenotypes estimated in step 1.

    Methods are described in the following publication:
        Sailer, Z. R. & Harms, M. J. 'Detecting High-Order Epistasis in
        Nonlinear Genotype-Phenotype Maps'. Genetics 205, 1079-1088 (2017).

    Parameters
    ----------
    model_type : str (default: global)
        type of epistasis model to use. See paper above for more information.

    Keyword Arguments
    -----------------
    Keyword arguments are interpreted as intial guesses for the nonlinear
    function parameters. Must have the same name as parameters in the
    nonlinear function

    Attributes
    ----------
    Additive : EpistasisLinearRegression
        Linear regression object for fitting additive model

    parameters : Parameters object
        Mapping object for nonlinear coefficients
    """
    def __init__(self, model_type="global", **p0):
        # Set up the function for fitting.
        self.function = power_transform
        self.minimizer = PowerTransformMinizer(**p0)
        self.parameters = self.minimizer.parameters
        self.order = 1
        self.Xbuilt = {}

        # Construct parameters object
        self.set_params(model_type=model_type)

        # Store model specs.
        self.model_specs = dict(
            function=self.function,
            model_type=self.model_type,
            **p0)

        # Set up additive and high-order linear model
        self.Additive = EpistasisLinearRegression(
            order=1, model_type=self.model_type)



# Mapping Object for epistatic interactions int the epistasis map
#
# Author: Zach Sailer
#
# ----------------------------------------------------------
# Outside imports
# ----------------------------------------------------------

import json
import itertools as it
from functools import wraps
from collections import OrderedDict

import numpy as np
import pandas as pd

# ----------------------------------------------------------
# Local imports
# ----------------------------------------------------------

import gpmap


def assert_epistasis(method):
    """Assert that an epistasis map has been attached to the object.
    """
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if hasattr(self, "epistasis") is False:
            raise AttributeError(
                self.__name__ + " does not an epistasis attribute set yet.")
        return method(self, *args, **kwargs)
    return wrapper


def site_to_key(site, state=""):
    """Convert site to key. `state` is added to end of key."""
    if type(state) != str:
        raise Exception("`state` must be a string.")
    return ",".join([str(l) for l in site]) + state


def key_to_site(key):
    """ Convert an interaction key to label."""
    return [int(k) for k in key.split(",")]


def genotype_coeffs(genotype, order=None):
    """List the possible epistatic coefficients (as label form) for a binary
    genotype up to a given order.
    """
    if order is None:
        order = len(genotype)
    length = len(genotype)
    mutations = [i + 1 for i in range(length) if genotype[i] == "1"]
    params = [[0]]
    for o in range(1, order + 1):
        params += [list(z) for z in it.combinations(mutations, o)]
    return params


def mutations_to_sites(order, mutations, start_order=0):
    """Build interaction sites up to nth order given a mutation alphabet.

    Parameters
    ----------
    order : int
        order of interactions
    mutations  : dict
        `mutations = { site_number : ["mutation-1", "mutation-2"] }`.
        If the site alphabet is note included, the model will assume binary
        between wildtype and derived.

    Example
    -------
    .. code-block:: python

        mutations = {
            0: ["A", "V"],
            1: ["A", "V"],
            ...
        }

    Returns
    -------
    sites : list
        list of all interaction sites for system with
        sequences of a given length and epistasis with given order.
    """
    # Convert a mutations mapping dictionary to a site mapping dictionary
    sitemap = dict()
    n_sites = 1
    for m in mutations:
        if mutations[m] is None:
            sitemap[m] = None
        else:
            sitemap[m] = list(range(n_sites, n_sites + len(mutations[m]) - 1))
            n_sites += len(mutations[m]) - 1

    # Include the intercept interaction?
    if start_order == 0:
        sites = [[0]]
        orders = range(1, order + 1)
    else:
        sites = list()
        orders = range(start_order, order + 1)

    length = len(sitemap)
    # Recursive algorithm that's difficult to follow.

    # Iterate through each order
    for o in orders:
        # Iterate through all combinations of orders with given length
        for term in it.combinations(range(length), o):
            # If any sites in `term` == None, skip this term.
            bad_term = False
            lists = []
            for i in range(len(term)):
                if sitemap[term[i]] is None:
                    bad_term = True
                    break
                else:
                    lists.append(sitemap[term[i]])
            # Else, add interactions combinations to list
            if bad_term is False:
                for r in it.product(*lists):
                    sites.append(list(r))

    return sites


class EpistasisMap(object):
    """DataFrame for epistatic interactions.
    """
    def __init__(self, sites, order=1, values=None, model_type="global"):
        self.order = order
        self.model_type = model_type
        data = {
            'sites': sites,
            'values': values,
        }
        self.data = pd.DataFrame(data)

    def map(self, attr1, attr2):
        """Dictionary that maps attr1 to attr2."""
        return dict(zip(getattr(self, attr1), getattr(self, attr2)))

    @classmethod
    def read_dataframe(cls, df, model_type="global"):
        """Create an epistasis model from dataframe"""
        sites = df['sites']
        values = df['values']
        order = max([len(site) for site in sites])
        self = cls(sites, order=order, values=values, model_type=model_type)
        return self

    def to_dict(self):
        """Get data as dictionary."""
        return self.data.to_dict('list')

    def to_csv(self, filename):
        """Write data to a csv file."""
        self.data.to_csv(filename)

    def to_excel(self, filename):
        """Write data to excel file."""
        self.data.to_excel(filename)

    @property
    def model_type(self):
        """Type of epistasis model used to get coefficients."""
        return self._model_type

    @property
    def n(self):
        """ Return the number of Interactions. """
        return len(self.data.sites)

    @property
    def order(self):
        """ Get order of epistasis in system. """
        return self._order

    @property
    def values(self):
        """ Get the values of the interaction in the system"""
        return self.data['values'].values

    @property
    def index(self):
        """ Get the interaction index in interaction matrix. """
        return self.data.index

    @property
    def sites(self):
        """ Get the interaction sites, which describe the position of
        interacting mutations in the genotypes. (type==list of lists,
        see self._build_interaction_sites)
        """
        return self.data.sites.values

    def get_orders(self, *orders):
        """Get epistasis of a given order."""
        return Orders(self, orders)

    # ----------------------------------------------
    # Setter Functions
    # ----------------------------------------------

    @order.setter
    def order(self, order):
        """"""
        self._order = order

    @values.setter
    def values(self, values):
        """ Manually set keys. NEED TO do some quality control here. """
        self.data.values = values

    @model_type.setter
    def model_type(self, model_type):
        types = ["global", "local"]
        if model_type not in types:
            raise Exception("Model type must be global or local")
        self._model_type = model_type


class Orders(map):
    """An object that provides API for easily calling epistasis of a given order
    in an epistasis map.
    """

    def __init__(self, epistasismap, orders):
        self._epistasismap = epistasismap
        self.orders = orders

    def __call__(self):
        """return a dictionary"""
        return dict(zip(self.keys, self.values))

    @property
    def df(self):
        """Dataframe for orders object."""
        data = {"sites": self.sites, "values": self.values,
                "stdeviations": self.stdeviations}
        return pd.DataFrame(data, columns=["sites", "values", "stdeviations"])

    @property
    def index(self):
        """Get indices of epistasis from this order."""
        # Check is multiple orders were given
        try:
            orders = list(iter(self.orders))
        except TypeError:
            orders = [self.orders]
        sites = self._epistasismap.sites
        x = [i for i in range(1, len(sites)) if len(sites[i]) in orders]
        # Add the zeroth element if included
        if 0 in orders:
            x = [0] + x
        return np.array(x)

    @property
    def sites(self):
        """Get epistatic sites"""
        return pd.Series([self._epistasismap.sites[int(i)]
                          for i in self.index], index=self.index)

    @property
    def values(self):
        """Get values of epistasis for this order."""
        return pd.Series([self._epistasismap.values[int(i)]
                          for i in self.index], index=self.index)



__doc__ = """Submodule with handy utilities used throughout the package.
"""
# -------------------------------------------------------
# Miscellaneous Python functions for random task
# -------------------------------------------------------

import abc
import itertools as it
import numpy as np
from scipy.misc import comb
from sklearn.metrics import mean_squared_error
from collections import OrderedDict

from gpmap.utils import genotypes_to_binary
from .mapping import mutations_to_sites

from epistasis.matrix import get_model_matrix


# -------------------------------------------------------
# Custom exceptions
# -------------------------------------------------------

class SubclassException(Exception):
    """ For methods that must be implemented in a subclass. """

# -------------------------------------------------------
# Useful methods
# -------------------------------------------------------

def genotypes_to_X(wildtype, genotypes,
    order=1,
    mutations=None,
    model_type='global'):
    """Build an X matrix for a list of genotypes."""
    # Binary representation
    binary = genotypes_to_binary(wildtype, genotypes, mutations)

    # Build list of sites from genotypes.
    sites = mutations_to_sites(order, mutations)

    # X matrix
    X = get_model_matrix(binary, sites, model_type=model_type)
    return X

# -------------------------------------------------------
# Useful Classes
# -------------------------------------------------------

class DocstringMeta(abc.ABCMeta):
    """Metaclass that allows docstring 'inheritance'

    Idea taken from this thread:
    https://github.com/sphinx-doc/sphinx/issues/3140
    """
    def __new__(mcls, classname, bases, cls_dict):
        # Create a new class as expected.
        cls = abc.ABCMeta.__new__(mcls, classname, bases, cls_dict)

        # Get order of inheritance
        mro = cls.__mro__[1:]

        # Iterate through items in class.
        for name, member in cls_dict.iteritems():

            # If the item does not have a docstring, add the base class docstring.
            if not getattr(member, '__doc__'):
                for base in mro:
                    try:
                        member.__doc__ = getattr(base, name).__doc__
                        break
                    except AttributeError:
                        pass
        return cls


class Bunch:
    """Classic bunch object for constructing empty objects."""

    def __init__(self, **kwds):
        self.__dict__.update(kwds)

    def update(self, **kwargs):
        """Turn a dictionary into an object with"""
        types = dict([(key, type(val)) for key, val in self.__dict__.items()])
        for key, value in kwargs.items():
            typed = types[key]
            if typed == np.ufunc:
                typed_val = value
            elif self.__dict__[key] is None:
                typed_val = value
            else:
                typed_val = types[key](value)
            setattr(self, key, typed_val)


def extract_mutations_from_genotypes(genotypes):
    """ Given a list of genotypes, infer a mutations dictionary.
    """
    genotypes_grid = [list(g) for g in genotypes]
    genotypes_array = np.array(genotypes_grid)
    (n_genotypes, n_sites) = genotypes_array.shape
    mutations = dict([(i, None) for i in range(n_sites)])
    for i in range(n_sites):
        unique = list(np.unique(genotypes_array[:, i]))
        if len(unique) != 1:
            mutations[i] = unique
    return mutations



# # External imports
# import unittest
# import pytest
#
# import numpy as np
# from gpmap import GenotypePhenotypeMap
#
# # Module to test
# from ..ensemble import EpistasisEnsembleRegression
#
#
# @pytest.fixture
# def gpm():
#     """Create a genotype-phenotype map"""
#     wildtype = "000"
#     genotypes = ["000", "001", "010", "100", "011", "101", "110", "111"]
#     phenotypes = [0.1,   0.1,   0.5,   0.4,   0.2,   0.8,   0.5,   1.0]
#     stdeviations = 0.1
#     return GenotypePhenotypeMap(wildtype, genotypes, phenotypes,
#                                 stdeviations=stdeviations)
#
#
# class TestEpistasisLinearRegression(object):
#
#     order = 3
#     nstates = 1
#
#     def test_init(self, gpm):
#         model = EpistasisEnsembleRegression(order=self.order, nstates=self.nstates)
#         model.add_gpm(gpm)
#
#         # Checks
#         check1 = model.order
#         check2 = model.model_type
#         assert check1 == self.order
#         assert check2 == "local"
#
#     def test_fit(self, gpm):
#         model = EpistasisEnsembleRegression(order=self.order, nstates=self.nstates)
#         model.add_gpm(gpm)
#         model.fit()
#         # Checks
#         check1 = hasattr(model, "Xbuilt")
#
#         # Tests
#         assert check1 is True
#         assert "obs" in model.Xbuilt
#         assert "fit" in model.Xbuilt
#
#     def test_fit_sample_weight(self, gpm):
#         model = EpistasisEnsembleRegression(order=self.order, nstates=self.nstates)
#         model.add_gpm(gpm)
#         model.fit(sample_weight='relative')
#
#     def test_predict(self, gpm):
#         model = EpistasisEnsembleRegression(order=self.order, nstates=self.nstates)
#         model.add_gpm(gpm)
#         model.fit()
#         check1 = model.predict()
#
#         # Tests
#         np.testing.assert_almost_equal(
#             sorted(check1), sorted(model.gpm.phenotypes))
#         assert "predict" in model.Xbuilt
#         assert "complete" in model.Xbuilt
#
#     def test_score(self, gpm):
#         model = EpistasisEnsembleRegression(order=self.order, nstates=self.nstates)
#         model.add_gpm(gpm)
#         model.fit()
#         score = model.score()
#         # Tests
#         assert score >= 0
#         assert score <= 1
#
#     # def test_hypothesis(self, gpm):
#     #     model = EpistasisEnsembleRegression(order=self.order, nstates=self.nstates)
#     #     model.add_gpm(gpm)
#     #     model.fit()
#     #     # Checks
#     #     check1 = model.hypothesis(thetas=model.coef_)
#     #     # Tests
#     #     np.testing.assert_almost_equal(
#     #         sorted(check1), sorted(model.gpm.phenotypes))
#
#     # def test_lnlikelihood(self, gpm):
#     #     model = EpistasisEnsembleRegression(order=self.order, nstates=self.nstates)
#     #     model.add_gpm(gpm)
#     #     model.fit()
#     #
#     #     # Calculate lnlikelihood
#     #     lnlike = model.lnlikelihood()
#     #     assert lnlike.dtype == float



import functools
import numpy as np

import matplotlib as mpl
import matplotlib.pyplot as plt


def plot_scale(
    model=None,
    yadd=None,
    yobs=None,
    yerr=None,
    function=None,
    cmap=None,
    color=None,
    s=50,
    ax=None,
    function_line=True,
    line_color='r',
    **kwargs):
    """Plot a Y_obs vs. Y_add showing the nonlinear scale in a genotype-phenotype
    map.

    Parameters
    ----------
    model : (default=None)
        Epistasis model.

    yadd : array (default=None)
        x-axis data. The additive model phenotypes.

    yobs : array (default=None)
        y-axis data. The observed phenotypes.

    yerr : array (default=None)
        y-axis error. Error in observed phenotypes.

    function : callable
        Nonlinear function.

    cmap : str
        Colormap name to map onto scatter points.

    color : str, array,
        color of phenotypes.

    s : int
        size of scatter points.

    ax : Axes
        Axes object to plot points on.

    function_line : bool
        If true, plots nonlinear function on top of points.

    line_color : matplotlib color.
        color of function line.


    Returns
    -------
    ax : matplotlib.Axes
        Axes object with plot.
    """
    if model is not None:
        params = model.parameters

        yobs = model.gpm.phenotypes
        yadd = model.Additive.predict()

        xx = np.linspace(min(yadd), max(yadd),20)
        yy = model.minimizer.predict(xx)

    elif yobs is None or yadd is None:
        raise Exception("If not model is given, pobs and padd must be set.")

    elif len(yobs) != len(yadd):
        raise Exception("pobs and padd must be same length.")

    # Set colormap and normalize colors.
    if cmap is not None:
        # Get colormap
        cmap_ = getattr(mpl.cm, cmap)
        cmap = cmap_(yobs)

        # Normalize the colors
        norm = mpl.colors.Normalize(vmin=min(yobs), vmax=max(yobs))
        color = mpl.cm.ScalarMappable(norm=norm, cmap=cmap_).to_rgba(yobs)

    # If no colormap, set color to black
    else:
        color = 'k'

    # Prepare plot
    if ax is None:
        fig, ax = plt.subplots()

    # Plot data.
    ax.scatter(yadd, yobs, color=color, s=s)

    # Plot errorbars if given
    if yerr is not None:
        ax.errorbar(yadd, yobs, yerr=yerr, fmt="none", ecolor='gray', zorder=-1)

    # Plot model
    if function_line:
        ax.plot(xx, yy, color=line_color, linewidth=2)
    return ax


@functools.wraps(plot_scale)
def plot_power_transform(
    model=None,
    yobs=None,
    yadd=None,
    function_line=True,
    function=None,
    line_color='r',
    *args,
    **kwargs):

    # Get data
    if model is not None:
        params = model.parameters

        yobs = model.gpm.phenotypes
        yadd = model.Additive.predict()
        xdata = yadd

        xx = np.linspace(min(yadd), max(yadd),20)
        yy = model.function(xx, **params, data=xdata)

    elif yobs is None or yadd is None or function is None:
        raise Exception("If not model is given, pobs and padd must be set.")

    elif len(yobs) != len(yadd):
        raise Exception("pobs and padd must be same length.")

    ax = plot_scale(
        model=model,
        yobs=yobs,
        yadd=yadd,
        function_line=False,
        function=function,
        *args, **kwargs,
    )

    # Plot model
    if function_line:
        ax.plot(xx, yy, color=line_color, linewidth=2)

    return ax



import pytest

# External imports
import numpy as np
from gpmap import GenotypePhenotypeMap

# Module to test
from ..classifiers import *


@pytest.fixture
def gpm():
    """Create a genotype-phenotype map"""
    wildtype = "000"
    genotypes = ["000", "001", "010", "100", "011", "101", "110", "111"]
    phenotypes = [0.0,   0.1,   0.5,   0.4,   0.2,   0.8,   0.5,   1.0]
    return GenotypePhenotypeMap(wildtype, genotypes, phenotypes)


class TestEpistasisLogisticRegression(object):

    # Set some initial parameters for this model
    order = 1
    threshold = 0.2

    def test_init(self, gpm):
        model = EpistasisLogisticRegression(threshold=self.threshold,
                                            model_type="local")
        model.add_gpm(gpm)
        # Tests
        assert model.order == self.order
        assert model.model_type == "local"

    def test_fit(self, gpm):
        model = EpistasisLogisticRegression(threshold=self.threshold,
                                            model_type="local")
        model.add_gpm(gpm)
        model.fit()

        assert hasattr(model, "classes")
        assert hasattr(model, "epistasis")
        assert hasattr(model, "coef_")

    def test_predict(self, gpm):
        model = EpistasisLogisticRegression(threshold=self.threshold,
                                            model_type="local")
        model.add_gpm(gpm)
        model.fit()
        ypred = model.predict()

        assert len(ypred) == model.gpm.n

    def test_predict_proba(self, gpm):
        model = EpistasisLogisticRegression(threshold=self.threshold,
                                            model_type="local")
        model.add_gpm(gpm)
        model.fit()
        probs = model.predict_proba()

        # check probs is the right length
        assert len(probs) == model.gpm.n

        # Check probs are between 0 and 1
        assert np.all(probs <= 1)
        assert np.all(probs >= 0)

    def test_predict_log_proba(self, gpm):
        model = EpistasisLogisticRegression(threshold=self.threshold,
                                            model_type="local")
        model.add_gpm(gpm)
        model.fit()
        probs = model.predict_log_proba()

        # check probs is the right length
        assert len(probs) == model.gpm.n

        # Check log probs are less than or equal to 0
        assert np.all(probs <= 0)

    def test_score(self, gpm):
        model = EpistasisLogisticRegression(threshold=self.threshold,
                                            model_type="local")
        model.add_gpm(gpm)
        model.fit()
        score = model.score()

        # Test score is between 0 and 1
        assert 0 <= score <= 1

    def test_thetas(self, gpm):
        model = EpistasisLogisticRegression(threshold=self.threshold,
                                            model_type="local")
        model.add_gpm(gpm)
        model.fit()

        # Check thetas is the correct length
        assert len(model.thetas) == len(model.coef_[0])

    def test_hypothesis(self, gpm):
        model = EpistasisLogisticRegression(threshold=self.threshold,
                                            model_type="local")
        model.add_gpm(gpm)
        model.fit()

        # these should be equal if working properly
        pred = model.predict_proba()[:, 0]
        hypo = model.hypothesis()
        np.testing.assert_almost_equal(pred, hypo)

    def test_lnlikelihood(self, gpm):
        model = EpistasisLogisticRegression(threshold=self.threshold,
                                            model_type="local")
        model.add_gpm(gpm)
        model.fit()
        lnlike = model.lnlikelihood()

        # Check we get a float
        assert lnlike.dtype == float



import json
import inspect
import numpy as np
import pandas as pd
from sklearn.preprocessing import binarize

from abc import abstractmethod, ABC, ABCMeta

# imports from gpmap dependency
from gpmap.gpm import GenotypePhenotypeMap

# Local imports
from epistasis.mapping import EpistasisMap, mutations_to_sites
from epistasis.matrix import get_model_matrix
from epistasis.utils import (extract_mutations_from_genotypes,
                             genotypes_to_X)
from .utils import XMatrixException
from sklearn.base import RegressorMixin, BaseEstimator

class SubclassException(Exception):
    """Subclass Exception for parent classes."""

def use_sklearn(sklearn_class):
    """Swap out base classes in an Epistasis model class with a sklearn_class +
    AbstractModel.
    """
    def mixer(cls):
        # Meta program the class
        name = cls.__name__
        methods = dict(cls.__dict__)

        # Put Sklearn first in line of parent classes
        parents = (sklearn_class, AbstractModel)

        # Rebuild class with Mixed in scikit learn.
        cls = type(name, parents, methods)
        return cls

    return mixer

class AbstractModel(ABC):
    """Abstract Base Class for all epistasis models.

    This class sets all docstrings not given in subclasses.
    """
    def __new__(self, *args, **kwargs):
        """Replace the docstrings of a subclass with docstrings in
        this base class.
        """
        # Get items in BaseModel.
        for name, member in inspect.getmembers(AbstractModel):
            # Get the docstring for this item
            doc = getattr(member, '__doc__')

            # Replace the docstring in self with basemodel docstring.
            try:
                member = getattr(self, name)
                member.__doc__ = doc

            except AttributeError:
                pass

        return super(AbstractModel, self).__new__(self)

    # --------------------------------------------------------------
    # Abstract Properties
    # --------------------------------------------------------------

    @property
    @abstractmethod
    def num_of_params(self):
        """Number of parameters in model.
        """
        raise SubclassException("Must be implemented in a subclass.")

    # --------------------------------------------------------------
    # Abstract Methods
    # --------------------------------------------------------------

    @abstractmethod
    def fit(self, X=None, y=None, **kwargs):
        """Fit model to data.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : None or ndarray (default=None)
            array of phenotypes. If None, the phenotypes in the attached
            genotype-phenotype map is used.


        Returns
        -------
        self :
            The model is returned. Allows chaining methods.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def fit_transform(self, X=None, y=None, **kwargs):
        """Fit model to data and transform output according to model.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : None or ndarray (default=None)
            array of phenotypes. If None, the phenotypes in the attached
            genotype-phenotype map is used.

        Returns
        -------
        gpm : GenotypePhenotypeMap
            The genotype-phenotype map object with transformed genotypes.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def predict(self, X=None):
        """Use model to predict phenotypes for a given list of genotypes.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        Returns
        -------
        y : ndarray
            array of phenotypes.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def predict_transform(self, X=None, y=None, **kwargs):
        """Transform a set of phenotypes according to the model.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        Returns
        -------
        y_transform : ndarray
            array of phenotypes.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def hypothesis(self, X=None, thetas=None):
        """Compute phenotypes from given model parameters.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        y : ndarray
            array of phenotypes predicted by model parameters.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def hypothesis_transform(self, X=None, y=None, thetas=None):
        """Transform phenotypes with given model parameters.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        y : ndarray
            array of phenotypes predicted by model parameters.
        """
        raise SubclassException("Must be implemented in a subclass.")


    @abstractmethod
    def lnlike_of_data(
           self,
           X=None,
           y=None,
           yerr=None,
           thetas=None):
        """Compute the individUal log-likelihoods for each datapoint from a set
        of model parameters.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        yerr : ndarray
            An array of the measured phenotypes standard deviations.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        y : ndarray
            array of phenotypes predicted by model parameters.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def lnlike_transform(
            self,
            X=None,
            y=None,
            yerr=None,
            lnprior=None,
            thetas=None):
        """Compute the individual log-likelihoods for each datapoint from a set
        of model parameters and a prior.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        yerr : ndarray
            An array of the measured phenotypes standard deviations.

        lnprior : ndarray
            An array of priors for a given datapoint.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        y : ndarray
            array of phenotypes predicted by model parameters.
        """
        raise SubclassException("Must be implemented in a subclass.")

    def lnlikelihood(
            self,
            X=None,
            y=None,
            yerr=None,
            thetas=None):
        """Compute the individal log-likelihoods for each datapoint from a set
        of model parameters.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        yerr : ndarray
            An array of the measured phenotypes standard deviations.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        lnlike : float
            log-likelihood of the model parameters.
        """
        lnlike = np.sum(self.lnlike_of_data(X=X, y=y, yerr=yerr, thetas=thetas))

        # If log-likelihood is infinite, set to negative infinity.
        if np.isinf(lnlike) or np.isnan(lnlike):
            return -np.inf
        return lnlike

    def add_X(self, X=None, key=None):
        """Add X to Xbuilt

        Keyword arguments for X:

        - None :
            Uses ``gpm.binary`` to construct X. If genotypes
            are missing they will not be included in fit. At the end of
            fitting, an epistasis map attribute is attached to the model
            class.


        Parameters
        ----------
        X :
            see above for details.
        key : str
            name for storing the matrix.

        Returns
        -------
        Xbuilt : numpy.ndarray
            newly built 2d array matrix
        """
        if isinstance(X, str) and X == None:

            if hasattr(self, "gpm") is False:
                raise XMatrixException("To build None, 'missing', or"
                                       "'complete' X matrix, a "
                                       "GenotypePhenotypeMap must be attached")

            # Get X columns
            columns = self.Xcolumns

            # Use desired set of genotypes for rows in X matrix.
            index = self.gpm.binary

            # Build numpy array
            x = get_model_matrix(index, columns, model_type=self.model_type)

            # Set matrix with given key.
            if key is None:
                key = X

            self.Xbuilt[key] = x

        elif type(X) == np.ndarray or type(X) == pd.DataFrame:
            # Set key
            if key is None:
                raise Exception("A key must be given to store.")

            # Store Xmatrix.
            self.Xbuilt[key] = X

        else:
            raise XMatrixException("X must be one of the following: None, "
                                   "'complete', numpy.ndarray, or "
                                   "pandas.DataFrame.")

        Xbuilt = self.Xbuilt[key]
        return Xbuilt

    def add_gpm(self, gpm):
        """Add a GenotypePhenotypeMap object to the epistasis model.
        """
        self._gpm = gpm

        # Reset Xbuilt.
        self.Xbuilt = {}

        # Construct columns for X matrix
        self.Xcolumns = mutations_to_sites(self.order, self.gpm.mutations)

        # Map those columns to epistastalis dataframe.
        self.epistasis = EpistasisMap(
            sites=self.Xcolumns,
            order=self.order,
            model_type=self.model_type)
        return self

    @property
    def gpm(self):
        """Data stored in a GenotypePhenotypeMap object."""
        return self._gpm

    # -----------------------------------------------------------
    # Argument handlers.
    # -----------------------------------------------------------

    def _X(self, data=None, method=None):
        """Handle the X argument in this model."""
        X = data
        # If X is None, see if we saved an array.
        if X is None:
            # Get X from genotypes
            X = genotypes_to_X(
                self.gpm.wildtype,
                self.gpm.genotypes,
                order=self.order,
                mutations=self.gpm.mutations,
                model_type=self.model_type
            )

        elif isinstance(X, str) and X in self.gpm.genotypes:
            # Get X from genotypes
            X = genotypes_to_X(
                self.gpm.wildtype,
                [X],
                order=self.order,
                mutations=self.gpm.mutations,
                model_type=self.model_type
            )

        # If X is a keyword in Xbuilt, use it.
        elif isinstance(X, str) and X in self.Xbuilt:
            X = self.Xbuilt[X]

        # If 2-d array, keep as so.
        elif isinstance(X, np.ndarray) and X.ndim == 2:
            pass

        # If list of genotypes.
        elif isinstance(X, list) or isinstance(X, np.ndarray):
            # Get X from genotypes
            X = genotypes_to_X(
                self.gpm.wildtype,
                X,
                order=self.order,
                mutations=self.gpm.mutations,
                model_type=self.model_type
            )
        else:
            raise Exception("X is invalid.")

        # Save X
        self.Xbuilt[method] = X
        return X

    def _y(self, data=None, method=None):
        """Handle y arguments in this model."""
        y = data
        if y is None:
            return self.gpm.phenotypes

        elif isinstance(y, np.ndarray) or isinstance(y, list):
            return y

        else:
            raise Exception("y is invalid.")

    def _yerr(self, data=None, method=None):
        """Handle yerr argument in this model."""
        yerr = data
        if yerr is None:
            return self.gpm.std.upper

        elif isinstance(yerr, np.ndarray) or isinstance(yerr, list):
            return yerr
        else:
            raise Exception("yerr is invalid.")

    def _thetas(self, data=None, method=None):
        """Handle yerr argument in this model."""
        thetas = data
        if thetas is None:
            return self.thetas

        elif isinstance(thetas, np.ndarray) or isinstance(thetas, list):
            return thetas
        else:
            raise Exception("thetas is invalid.")

    def _lnprior(self, data=None, method=None):
        _lnprior = data
        if _lnprior is None:
            return np.zeros(self.gpm.n)

        elif isinstance(_lnprior, np.ndarray) or isinstance(_lnprior, list):
            return _lnprior
        else:
            raise Exceptison("_prior is invalid.")


class BaseModel(AbstractModel, RegressorMixin, BaseEstimator):
    """Base model for defining an epistasis model.
    """
    pass



# import numpy as np
# import pytest
#
# from ..power import PowerScaleSimulation
#
# class TestPowerScaleSimulation(object):
#
#     wildtype = "00"
#     mutations = {0:["0","1"], 1:["0","1"]}
#     p0 = [1,1,1]
#
#     def test_init(self):
#         sim = PowerScaleSimulation(self.wildtype, self.mutations, p0=self.p0)
#         # Checks
#         assert isinstance(sim, PowerScaleSimulation)
#         assert hasattr(sim, "Xbuilt")
#
#     def test_function(self):
#         x = np.random.uniform(1,5, size=10)
#
#         # Run function, catch output
#         output = PowerScaleSimulation.function(x, *self.p0)
#
#         # Assert callable
#         assert callable(PowerScaleSimulation.function)
#         # assert type(output) = np.ndarray
#         assert len(output) == len(x)
#
#     def test_build(self):
#         sim = PowerScaleSimulation(self.wildtype, self.mutations, p0=self.p0)
#         sim.set_coefs_order(2)
#         sim.set_coefs_decay()
#
#         assert hasattr(sim, "phenotypes")
#         assert hasattr(sim, "linear_phenotypes")
#         assert "complete" in sim.Xbuilt



import numpy as np
import pandas as pd
import lmfit

from sklearn.base import BaseEstimator, RegressorMixin


from epistasis.stats import pearson
from ..mapping import EpistasisMap, mutations_to_sites
from .base import BaseModel
from epistasis.matrix import get_model_matrix
from .utils import arghandler


class State(EpistasisMap):
    """A state in an EpistasisEnsembleModel.

    This represents/models an unknown excited state that is perturbed by
    mutations in the genotype-phenotype map and contributes to the overall
    phenotype.

    Parameters
    ----------
    name : string
        name of the state. Should have the form 'state_X' where X is a letter.

    sites : list of lists
        each element in list is a list of genotype site indexes. These sites
        are additive/epistatic coefficients contributing to this state.

    Attributes
    ----------
    See EpistasisMap for more details.

    keys : list
        list of lmfit.Parameter keys.
    """
    def __init__(self, name, sites, *args, **kwargs):
        # Call super init.
        super(State, self).__init__(sites=sites, *args, **kwargs)

        # Set name.
        self.name = name

        # Construct parameters object
        self.parameters = lmfit.Parameters()

        # fill parameters .
        for key in self.keys:
            self.parameters.add(key, value=0)

    @property
    def keys(self):
        """Coefficient Parameter keys."""
        keys = []
        for sites in self.sites:
            key = "".join([str(ch) for ch in sites])
            name = "{}_{}".format(self.name, key)
            keys.append(name)
        return keys


class EpistasisEnsembleRegression(BaseModel):
    """Ensemble epistasis model. It models variation in a genotype-phenotype map
    as a statistical ensemble of nstates contributing to each genotype's
    phenotype.

    Attributes
    ----------
    See BaseModel for more details.

    parameters : lmfit.Parameters
        Parameters resulting from fit.
    """

    _ALPHABET = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J']

    def __init__(self, order=1, nstates=2):
        self.nstates = nstates
        self.model_type = 'local'
        self.order = order
        self.states = {}
        self.Xbuilt = {}
        self.parameters = lmfit.Parameters()

    def add_gpm(self, gpm):
        """Add genotype-phenotype map to model object."""
        super(EpistasisEnsembleRegression, self).add_gpm(gpm)

        # Add states to model.
        for i in range(self.nstates):
            # State name
            name = "state_{}".format(self._ALPHABET[i])

            # Add state.
            self.add_state(name)

        return self

    def add_state(self, name):
        """ Add a state to the model."""
        sites = self.Xcolumns

        # Create state.
        state = State(name, sites, model_type=self.model_type)

        # Store state.
        self.states[name] = state

        # Set as attribute.
        setattr(self, name, state)

        return self

    @property
    def num_of_params(self):
        """Return number of parameters in model."""
        n = 0
        n += len(self.parameters)
        return n

    @property
    def parameters(self):
        """All parameters in the value."""
        parameters = lmfit.Parameters()

        # Get parameter data.
        for state in self.states.values():
            parameters.add_many(*state.parameters.values())

        return parameters

    @parameters.setter
    def parameters(self, parameters):
        """Set parameters for all states."""

        # Add state parameters.
        for state in self.states.values():
            keys = state.keys
            parameters_ = lmfit.Parameters()
            values = []
            for key in state.keys:
                p = parameters[key]
                parameters_.add_many(p)
                values.append(p.value)

            parameters_.add_many(*[parameters[key] for key in state.keys])
            state.parameters = parameters_
            state.values = values

    def functional_form(self, thetas, X=None):
        """Ensemble function calculating phenotypes using a Boltzmann weighted
        ensemble of states, where each state is constructed from linear
        epistasis models.

        Parameters
        ----------
        thetas : array
            array of parameters values ordered alphabetically.

        X : 2d-array (default None)
            X matrix to use for linear portion of models. If None, uses the
            matrix stored under the 'fit' key.
        """
        length = self.states['state_A'].n
        nstates = len(self.states)

        if X is None:
            X = self.Xbuilt['fit']

        # Calculate a partition function
        Z = 0
        for state_i in range(nstates):
            # Get parameter indexes
            idx_start = state_i * length
            idx_stop = state_i * length + length
            dDG = thetas[idx_start:idx_stop]

            # Get state.
            state = self.states['state_{}'.format(self._ALPHABET[state_i])]

            # Additive model.
            additive = X @ dDG

            # add to ensemble
            Z += np.exp(-additive)

        # Ensemble model.
        y = np.log(Z)
        return y

    @arghandler
    def fit(self, X=None, y=None, **kwargs):
        """Fit ensemble model to data.

        Parameters
        ----------
        X : 2-d array
            independent data; samples.
        y : array
            dependent data; observations.
        """
        # X matrix.
        self.add_X(X=X, key='fit')

        # Storing failed residuals
        last_residual_set = None

        # Residual function to minimize.
        def residual(params, func, y=None):
            # Fit model
            parvals = list(params.values())
            ymodel = func(parvals)

            # Store items in case of error.
            nonlocal last_residual_set
            last_residual_set = (params, ymodel)

            return y - ymodel

        y = self.gpm.phenotypes

        # Minimize the above residual function.
        self.results = lmfit.minimize(
            residual, self.parameters,
            args=[self.functional_form],
            kws={'y': y})

        # Set parameters fitted by model.
        self.parameters = self.results.params

        return self

    def fit_transform(self, X=None, y=None, **kwargs):
        """Same as calling fit in ensemble model.
        """
        self.fit(X=X, y=y, **kwargs)

    @arghandler
    def predict(self, X=None):
        """Predict phenotypes using fitted model.
        """
        return self.functional_form(list(self.parameters.values()), X=X)

    def predict_transform(self, X=None, y=None):
        """Same as calling predict."""
        return self.predict(X=X)

    @arghandler
    def score(self, X=None, y=None, **kwargs):
        """Calculate the pearson coefficient between the models predictions and
        a given y array.
        """
        return pearson(self.gpm.phenotypes, self.predict(X=X))**2



import numpy as np
import pandas as pd

# Scikit-learn classifiers
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
from sklearn.naive_bayes import BernoulliNB
from sklearn.preprocessing import binarize

from ..mapping import EpistasisMap
from .base import BaseModel, use_sklearn
from .utils import (XMatrixException, arghandler)

from .linear import EpistasisLinearRegression

from gpmap import GenotypePhenotypeMap

# Suppress Deprecation warning
import warnings
warnings.filterwarnings(action="ignore", module="sklearn",
                        category=DeprecationWarning)


@use_sklearn(LogisticRegression)
class EpistasisLogisticRegression(BaseModel):
    """Logistic regression for estimating epistatic interactions that lead to
    nonviable phenotypes. Useful for predicting viable/nonviable phenotypes.

    Parameters
    ----------
    threshold : float
        value below which phenotypes are considered nonviable.

    order : int
        order of epistasis model

    model_type : str (default="global")
        type of model matrix to use. "global" defines epistasis with respect to
        a background-averaged "genotype-phenotype". "local" defines epistasis
        with respect to the wildtype genotype.
    """
    def __init__(self, threshold, model_type="global", **kwargs):
        super(self.__class__, self).__init__(**kwargs)
        self.threshold = threshold
        self.model_type = model_type
        self.fit_intercept = False
        self.order = 1
        self.Xbuilt = {}

        # Store model specs.
        self.model_specs = dict(
            threshold=self.threshold,
            model_type=self.model_type,
            **kwargs)

        # Set up additive linear model for pre-classifying
        self.Additive = EpistasisLinearRegression(
            order=1, model_type=self.model_type)

    def fit(self, X=None, y=None, **kwargs):
        # Use Additive model to establish the phenotypic scale.
        # Prepare Additive model
        self.Additive.add_gpm(self.gpm)

        # Prepare a high-order model
        self.Additive.epistasis = EpistasisMap(
            sites=self.Additive.Xcolumns,
            order=self.Additive.order,
            model_type=self.Additive.model_type
        )

        # Fit the additive model and infer additive phenotypes
        self.Additive.fit(X=X, y=y)
        padd = self.Additive.predict(X=X)
        self = self._fit_(X=X, y=y)
        return self

    def fit_transform(self, X=None, y=None, **kwargs):
        self.fit(X=X, y=y, **kwargs)
        ypred = self.predict(X=X)


        # Transform map.
        gpm = GenotypePhenotypeMap.read_dataframe(
            dataframe=self.gpm.data[ypred==1],
            wildtype=self.gpm.wildtype,
            mutations=self.gpm.mutations
        )
        return gpm

    @property
    def num_of_params(self):
        n = 0
        n += self.epistasis.n
        return n

    @arghandler
    def _fit_(self, X=None, y=None, **kwargs):
        # Fit the classifier
        yclass = binarize(y.reshape(1, -1), self.threshold)[0]
        self.classes = yclass
        super(self.__class__, self).fit(X=X, y=yclass, **kwargs)
        self.epistasis.values = np.reshape(self.coef_, (-1,))
        return self

    @arghandler
    def predict(self, X=None):
        return super(self.__class__, self).predict(X=X)

    def predict_transform(self, X=None, y=None):
        x = self.predict(X=X)
        y[x <= 0.5] = self.threshold
        return y

    @arghandler
    def predict_log_proba(self, X=None):
        return super(self.__class__, self).predict_log_proba(X)

    @arghandler
    def predict_proba(self, X=None):
        return super(self.__class__, self).predict_proba(X)

    @arghandler
    def score(self, X=None, y=None, **kwargs):
        yclass = binarize(y.reshape(1, -1), threshold=self.threshold)[0]
        return super(self.__class__, self).score(X=X, y=yclass)

    @arghandler
    def lnlike_of_data(self, X=None, y=None, yerr=None, thetas=None):
        # Calculate Y's
        ymodel = self.hypothesis(X=X, thetas=thetas)
        ymodel_ = 1 - ymodel
        ymodel[ymodel < 0.5] = ymodel_[ymodel < 0.5]

        return np.log(ymodel)

    @arghandler
    def lnlike_transform(
            self,
            X=None,
            y=None,
            yerr=None,
            lnprior=None,
            thetas=None):
        # Update likelihood.
        ymodel = self.hypothesis(X=X, thetas=thetas)
        yclass = np.ones(len(ymodel))
        yclass[ymodel > 0.5] = 0

        lnlike = self.lnlike_of_data(X=X, y=y, yerr=yerr, thetas=thetas)
        lnprior[yclass == 0] = 0
        return lnlike + lnprior

    @arghandler
    def hypothesis(self, X=None, thetas=None):
        # Calculate probability of each class
        logit_p0 = 1 / (1 + np.exp(np.dot(X, thetas)))

        # Returns probability of class 1
        return logit_p0

    def hypothesis_transform(self, X=None, y=None, thetas=None):
        ypred = self.hypothesis(X=X, thetas=thetas)
        y[ypred > 0.5] = self.threshold
        return y

    @property
    def thetas(self):
        return self.epistasis.values



import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as patches
import matplotlib as mpl
import numpy as np
import gpmap
from scipy.stats import norm as scipy_norm
from epistasis.utils import Bunch


def plot_coefs(model=None, sites=None, values=None, errors=None, **kwargs):
    """Create a barplot with the values from model, drawing the x-axis as a
    grid of boxes indicating the coordinate of the epistatic parameter.
    Should automatically generate an almost publication-quality figure.

    Parameters
    ----------
    model: BaseModel object
        epistasis model.

    sites : array
        array of epistatic indices/sites.

    values : array
        an array of epistatic coefficients

    errors : 2d array or list
        upper and lower bounds for each beta.

    Keyword arguments
    -----------------
    logbase : numpy.ufunc (default=np.log10)
        function to transform into log space

    log_transform : bool (default=False)
        transform the values if true.

    order_colors :
        list/tuple of colors for each order (rgb,html string-like)

    significance :
        how to treat signifiance.  should be
        1. "bon" -> Bonferroni corrected p-values (default)
        2. "p" -> raw p-values
        3. None -> ignore significance

    significance_cutoff :
        value above which to consider a term significant

    sigmas :
        number of sigmas to show for each error bar

    y_scalar :
        how much to scale the y-axis above and beyond y-max

    y_axis_name :
        what to put on the y-axis of the barplot

    figsize :
        tuple of figure width,height

    height_ratio :
        how much to scale barplot relative to xbox

    star_cutoffs :
        signifiance cutoffs for star stack.  should go from highest
                  p to lowest p (least to most significant)

    star_spacer :
        constant that scales how closely stacked stars are from one
        another

    ybounds : tuple (default=None)

    bar_borders : bool (default=True)

    xgrid : bool (default=True)

    ecolor : color (default='black')

    elinewidth : float (default=1)

    capthick : float (default=1)

    capsize : float (default=1)

    gridlines : float (default=1)
        x grid linewidth

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object

    ax : matplotlib.pyplot.Axes
        Axes object
    """
    # Some sanity checks.
    if model is not None:
        sites = model.epistasis.sites
        values = model.epistasis.values

    else:
        if sites is None:
            raise Exception("If no model is given, sites and values must be "
                            "set.")
        elif len(sites) != len(values):
            raise Exception("sites and values must be the same length.")

    # Set up plotting user options. Type check the options to make sure nothing
    # will break. Also helps with widgets.
    sites = list(sites)
    # Prepare an cycle of colors
    order = len(sites[-1:])
    prop_cycle = plt.rcParams['axes.prop_cycle']
    color_cycle = prop_cycle.by_key()['color']
    color_scalar = int(order / len(color_cycle)) + 1
    color_cycle *= color_scalar

    defaults = {
        "order_colors": color_cycle,
        "logbase": np.log10,
        "log_transform": False,
        "significance": "bon",
        "significance_cutoff": 0.05,
        "sigmas": 0,
        "log_space": False,
        "y_scalar": 1.5,
        "y_axis_name": "",
        "figwidth": 5,
        "figheight": 3,
        "figsize": (5, 3),
        "height_ratio": 12,
        "star_cutoffs": (0.05, 0.01, 0.001),
        "star_spacer": 0.0075,
        "ybounds": None,
        "bar_borders": True,
        "xgrid": True,
        "ecolor": "black",
        "capthick": 1,
        "capsize": 1,
        "elinewidth": 1,
        "save": False,
        "fname": "figure.svg",
        "format": "svg",
        "gridlines": 1,
    }
    # types = dict([(key, type(val)) for key, val in defaults.items()])
    # defaults.update(kwargs)
    # options = objectify(defaults)
    options = Bunch(**defaults)
    options.update(**kwargs)
    # Construct keyword arguments
    error_kw = {
        "ecolor": options.ecolor,
        "capsize": options.capsize,
        "elinewidth": options.elinewidth,
        "capthick": options.capthick,
    }
    if "figsize" in kwargs:
        options.figsize = kwargs["figsize"]
    else:
        options.figsize = (options.figwidth, options.figheight)

    # Name all variables that matter for this function
    if sites[0] == [0]:
        sites = sites[1:]
        values = values[1:]
        if errors is not None:
            upper = errors[1][1:]
            lower = errors[0][1:]
    else:
        if errors is not None:
            upper = errors[1]
            lower = errors[0]

    # Sanity check on the errors
    if options.sigmas == 0:
        significance = None
    elif options.significance is None:
        sigmas = 0

    # Figure out the length of the x-axis and the highest epistasis observed
    num_terms = len(sites)
    highest_order = max([len(l) for l in sites])

    # Figure out how many sites are in the dataset (in case of non-binary
    # system)
    all_sites = []
    for l in sites:
        all_sites.extend(l)
    all_sites = list(dict([(s, []) for s in all_sites]).keys())
    all_sites.sort()
    num_sites = len(all_sites)

    # Figure out how to color each order
    if options.order_colors is None:
        options.order_colors = ["gray" for i in range(highest_order + 1)]
    else:
        if len(options.order_colors) < highest_order:
            raise ValueError("order_colors has too few entries "
                             "(at least {:d} needed)\n".format(highest_order))

        # Stick gray in the 0 position for insignificant values
        options.order_colors = list(options.order_colors)
        options.order_colors.insert(0, "gray")

    # ---------------------- #
    # Deal with significance #
    # ---------------------- #
    # NEED TO RETURN TO SIGNIFICANCE FUNCTIONS
    if options.sigmas == 0:
        options.significance = None
    else:
        # If log transformed, need to get raw values for normal distribution
        if options.log_transform:
            z_score = abs((values - 1) / upper)
        # else, just grab standard values
        else:
            z_score = abs((values) / upper)

        # if z_score is > 5, set z_score to largest possible range
        # where p-value is within floating point
        z_score[z_score > 8.2] = 8.2

    # straight p-values
    if options.significance == "p":
        p_values = 2 * (1 - scipy_norm.cdf(z_score))

    # bonferroni corrected p-values
    elif options.significance == "bon":
        p_values = 2 * (1 - scipy_norm.cdf(z_score)) * len(values)

    # ignore p-values and color everything
    elif options.significance is None:
        p_values = [0 for i in range(len(sites))]
        options.significance_cutoff = 1.0

    # or die
    else:
        raise ValueError("signifiance argument {:s} not "
                         "recognized\n".format(options.significance))

    # Create color array based on significance
    color_array = np.zeros((len(sites)), dtype=int)
    for i, l in enumerate(sites):
        if p_values[i] < options.significance_cutoff:
            color_array[i] = len(l) - 1
        else:
            color_array[i] = -1

    # ---------------- #
    # Create the plots #
    # ---------------- #

    # Make a color map
    cmap = mpl.colors.ListedColormap(colors=options.order_colors)
    # set the 'bad' values (nan) to be white and transparent
    cmap.set_bad(color='w', alpha=0)
    bounds = range(-1, len(options.order_colors))
    norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

    if options.xgrid is True:
        fig = plt.figure(figsize=options.figsize)

        n_coefs = len(sites)
        n_sites = max([max(l) for l in sites])

        # Calculate the height_ratio of the grid and the bar graph
        box_size = options.figsize[0] / float(n_coefs)
        grid_height = box_size * n_sites
        bar_height = options.figsize[1] - grid_height
        height_ratio = bar_height / grid_height

        # Create a plot with an upper and lower panel, sharing the x-axis
        gs = mpl.gridspec.GridSpec(2, 1,
                                   height_ratios=[height_ratio, 1],
                                   hspace=0.00)

        ax = [plt.subplot(gs[0])]
        ax.append(plt.subplot(gs[1], sharex=ax[0]))
        bar_axis = ax[0]
        grid_axis = ax[1]
        # Create the box-array x-axis
        # path codes for drawing the boxes
        box_codes = [Path.MOVETO,
                     Path.LINETO,
                     Path.LINETO,
                     Path.LINETO,
                     Path.CLOSEPOLY]

        color_vector = options.order_colors
        for i in range(n_coefs):
            for j in range(n_sites):
                color = "None"
                if j + 1 in sites[i]:
                    color = color_vector[len(sites[i])]
                # vertices for a given square
                verts = [
                    (i, n_coefs - j),
                    (i, n_coefs - j - 1),
                    (i + 1, n_coefs - j - 1),
                    (i + 1, n_coefs - j),
                    (i, n_coefs - j),
                ]
                # Create a patch for a square
                path = Path(verts, box_codes)
                patch = patches.PathPatch(path,
                                          facecolor=color,
                                          lw=options.gridlines)
                grid_axis.add_patch(patch)

        grid_axis.axis('equal')
        grid_axis.axis('off')

    else:

        fig, ax = plt.subplots(figsize=options.figsize)
        bar_axis = ax

    # ------------------ #
    # Create the barplot #
    # ------------------ #

    # set up bar colors
    # prop_cycle = plt.rcParams['axes.prop_cycle']
    # colors_for_bar = prop_cycle.by_key()['color']
    colors_for_bar = np.array([mpl.colors.colorConverter.to_rgba(
        options.order_colors[(i + 1)]) for i in color_array])

    # Plot without errors
    if options.sigmas == 0:
        if options.log_space:
            bar_y = options.logbase(values)
        else:
            bar_y = values
        bar_axis.bar(np.arange(len(bar_y)) + .55, bar_y, width=.9,
                     color=colors_for_bar, edgecolor="none")
    # plot with errors
    else:
        bar_y = values
        upper = options.sigmas * upper
        lower = options.sigmas * lower       # Plot the graph on a log scale
        if options.log_space:
            new_bar_y = options.logbase(bar_y)
            new_upper = gpmap.errors.upper_transform(bar_y, upper,
                                                     options.logbase)
            new_lower = gpmap.errors.lower_transform(bar_y, lower,
                                                     options.logbase)
        # else if the space is log transformed,
        # plot the non-log interaction values
        else:
            new_upper = upper
            new_lower = lower
            new_bar_y = bar_y
        yerr = [new_lower, new_upper]
        # Plot
        bar_axis.bar(np.arange(len(bar_y)) + 0.05, new_bar_y,
                     width=0.9,
                     yerr=yerr,
                     color=colors_for_bar,
                     error_kw=error_kw,
                     edgecolor="none",
                     linewidth=2)
    # Add horizontal lines for each order
    bar_axis.hlines(0, 0, len(values), linewidth=1, linestyle="-", zorder=0)
    # Label barplot y-axis
    bar_axis.set_ylabel(options.y_axis_name, fontsize=14)
    # Set barplot y-scale
    if options.ybounds is None:
        ymin = -options.y_scalar * max(abs(bar_y))
        ymax = options.y_scalar * max(abs(bar_y))
    else:
        ymin = options.ybounds[0]
        ymax = options.ybounds[1]

    # Make axes pretty pretty
    bar_axis.axis([-1, len(bar_y) + 1, ymin, ymax])
    bar_axis.set_frame_on(False)  # axis("off")
    bar_axis.get_xaxis().set_visible(False)
    bar_axis.get_yaxis().tick_left()
    bar_axis.get_yaxis().set_tick_params(direction='out')
    bar_axis.add_artist(mpl.lines.Line2D((-1, -1),
                        (bar_axis.get_yticks()
                        [1], bar_axis.get_yticks()[-2]),
                        color='black', linewidth=1))

    # add vertical lines between order breaks
    previous_order = 1
    for i in range(len(sites)):
        if len(sites[i]) != previous_order:
            bar_axis.add_artist(mpl.lines.Line2D((i, i),
                                                 (ymin, ymax),
                                                 color="black",
                                                 linestyle=":",
                                                 linewidth=1))
            previous_order = len(sites[i])

    # ------------------------- #
    # Create significance stars #
    # ------------------------- #
    if options.sigmas != 0:
        min_offset = options.star_spacer * (ymax - ymin)
        for i in range(len(p_values)):

            star_counter = 0
            for j in range(len(options.star_cutoffs)):
                if p_values[i] < options.star_cutoffs[j]:
                    star_counter += 1
                else:
                    break

            for j in range(star_counter):
                bar_axis.text(x=(i + 0),
                              y=ymin + (j * min_offset),
                              s="*", fontsize=16)

    # remove x tick labels
    try:
        plt.setp([a.get_xticklabels() for a in fig.axes[:-1]], visible=False)
    except IndexError:
        pass

    # Draw the final figure
    # fig.tight_layout()

    if options.save:
        fig.savefig(options.fname, format=options.format)

    return fig, ax



import inspect
import lmfit
from lmfit import Parameter, Parameters

from abc import ABC, abstractmethod

# ---------------------- Abstract Minimizer class -------------------------

class Minimizer(ABC):
    """Abstract Base Class for function minimizers.

    Wraps a function. Fits data using the function. Then, that fitted function
    can be used to transform data.
    """
    def __init__(self, function, **p0):
        pass

    @abstractmethod
    def function(self, x, **params):
        """Execute function given parameters.
        """

    @abstractmethod
    def predict(self, x):
        """Predict y, given x, from the fitted model.
        """

    @abstractmethod
    def transform(self, x, y):
        """Method that transforms y onto the x scale using the minimized fit.
        """

    @abstractmethod
    def fit(self, x, y):
        """Fit the x, y data using the function. Store fit.
        """

# ---------------------- Abstract Minimizer class -------------------------


class FunctionMinimizer(Minimizer):
    """Minimizer object that fits data using a function.
    """
    def __init__(self, function, **p0):
        # Do some inspection to get the parameters from the nonlinear
        # function argument list.
        func_signature = inspect.signature(function)
        func_params = list(func_signature.parameters.keys())

        if func_params[0] != "x":
            raise Exception("First argument of the nonlinear function must "
                            "be `x`.")

        # Construct lmfit parameters object
        self.parameters = Parameters()
        for p in func_params[1:]:
            # Get starting value of parameter if given.
            val = None
            if p in p0:
                val = p0[p]
            # Add parameter.
            self.parameters.add(name=p, value=val)

        # Set function
        self._function = function

    def function(self, x, *args, **kwargs):
        """Execute the function."""
        return self._function(x, *args, **kwargs)

    def predict(self, x):
        """Call function"""
        return self._function(x, **self.parameters)

    def transform(self, x, y):
        """Transform y onto the x scale."""
        ymodel = self.predict(x)
        return (y - ymodel) + x

    def fit(self, x, y):
        """Fit the function"""
        # Store residual steps in case fit fails.
        last_residual_set = None

        # Residual function to minimize.
        def residual(params, func, x, y=None):
            # Fit model
            parvals = list(params.values())
            ymodel = func(x, *parvals)

            # Store items in case of error.
            nonlocal last_residual_set
            last_residual_set = (params, ymodel)

            return y - ymodel

        # Minimize the above residual function.
        try:
            self.minimizer = lmfit.minimize(
                residual,
                self.parameters,
                args=[self._function, x],
                kws={'y': y})

            # Point to nonlinear.
            self.parameters = self.minimizer.params

        # If fitting fails, print what happened
        except Exception as e:
            # if e is ValueError
            print("ERROR! Some of the transformed phenotypes are invalid.")
            print("\nParameters:")
            print("----------")
            print(last_residual_set[0].pretty_print())
            print("\nTransformed phenotypes:")
            print("----------------------")
            print(last_residual_set[1])
            raise e



import numpy as np

# Try importing model matrix builder from cython extension for speed up.
try:
    from .matrix_cython import build_model_matrix

except ImportError:

    import warnings as _warnings

    # Raise warning
    _warnings.warn('Could not load cython extension, "build_model_matrix".', Warning)

    def build_model_matrix(encoding_vectors, sites):
        """Build model matrix.

        This can be made faster by cython.
        """
        # get dimensions of matrix
        n, m = len(encoding_vectors), len(sites)

        matrix = np.ones((n,m), dtype=int)

        # Interate over rows
        for i in range(n):
            vec = encoding_vectors[i]

            # Iterate over cols
            for j in range(m):
                # Get sites in this coefficient column
                players = sites[j]

                # Multiply these elements
                matrix[i,j] = np.prod(vec[players])

        return matrix


def encode_vectors(binary_genotypes, model_type='global'):
    """Encode a set of binary genotypes is input vectors for the given model

    """
    # Initialize vector container
    vectors = []

    # Handle a global model
    if model_type == 'global':

        for i, genotype in enumerate(binary_genotypes):
            vector = np.array([0] + list(genotype), dtype=float)
            vector[vector==1] = -1
            vector[vector==0] = 1
            vectors.append(vector)

    # Handle a local model.
    elif model_type == 'local':

        for i, genotype in enumerate(binary_genotypes):
            vector = np.array([1] + list(genotype), dtype=float)
            vectors.append(vector)

    # Don't understand the model
    else:
        Exception("Unrecognized model type.")

    return np.array(vectors)


def get_model_matrix(binary_genotypes, sites, model_type='global'):
    """Get a model matrix for a given set of genotypes and coefficients.
    """
    # Convert sites to array of arrays
    sites = np.array([np.array(s) for s in sites])

    # Encode genotypes
    encoded_vector = encode_vectors(binary_genotypes, model_type=model_type)

    # Build matrix.
    X = build_model_matrix(encoded_vector, sites)
    return X



# import lmfit parameters
from .minimizer import Parameter, Parameters

# Import models
from .ordinary import EpistasisNonlinearRegression
from .power import EpistasisPowerTransform
from .spline import EpistasisSpline



from distutils.core import setup
from Cython.Build import cythonize
import numpy

# To build extension, run python setup.py build_ext --inplace
setup(
    ext_modules=cythonize("matrix_cython.pyx"),
    include_dirs=[numpy.get_include()]
)



# #from gpmap.simulate import GenotypePhenotypeSimulation
# from ..utils import *
#
# from ..base import BaseModel
#
# class MockModel(BaseModel):
#
#     def __init__(self):
#         self.gpm = GenotypePhenotypeSimulation.from_length(2)
#         self.model_type = "local"
#         self.order = 2
#         self.Xbuilt = {}
#
#     @X_fitter
#     def fit(self, X='obs', y='obs'):
#         self.coef_ = [0,0,0,0]
#         return None
#
#     @X_predictor
#     def predict(self, X='complete', y='complete'):
#         return None
#
# def test_X_fitter():
#     model = MockModel()
#     model.fit()
#     # Test an Xfit matrix was made
#     assert "obs" in model.Xbuilt
#     assert "fit" in model.Xbuilt
#     assert  model.Xbuilt["fit"].shape == (4,4)
#
# def test_X_predictor():
#     model = MockModel()
#     model.fit()
#     model.predict()
#     # Test an Xfit matrix was made
#     assert "complete" in model.Xbuilt
#     assert "predict" in model.Xbuilt
#     assert model.Xbuilt["predict"].shape == (4,4)



"""
A library of models to decompose high-order epistasis in genotype-phenotype
maps.
"""
from .linear import (EpistasisLinearRegression,
                     EpistasisLasso,
                     EpistasisRidge,
                     EpistasisElasticNet)
from .nonlinear import (EpistasisNonlinearRegression,
                        EpistasisPowerTransform,
                        EpistasisSpline)
from .classifiers import EpistasisLogisticRegression
from .pipeline import EpistasisPipeline



__doc__ = """Submodule with useful statistics functions for epistasis model."""

# -----------------------------------------------------------------------
# Useful statistical metrics as methods
# -----------------------------------------------------------------------

import numpy as np
from scipy.stats import f
from scipy.stats import norm
import scipy

from gpmap import GenotypePhenotypeMap

# -----------------------------------------------------------------------
# Correlation metrics
# -----------------------------------------------------------------------

def split_data(data, fraction=1.0):
    """Split DataFrame into two sets, a training and a test set.

    Parameters
    ----------
    data : pandas.DataFrame
        full dataset to split.

    fraction : float
        fraction in training set.

    Returns
    -------
    train_set : pandas.DataFrame
        training set.

    test_set : pandas.DataFrame
        test set.
    """
    if  0 < fraction > 1.0:
        raise Exception("fraction is invalid.")

    length = len(data)
    n = int(length * fraction)
    frac = n / length

    # Shuffle the indices
    index = np.arange(0, length, dtype=int)
    np.random.shuffle(index)

    train_idx = index[:n]
    test_idx = index[n:]

    # Split data.
    train_set = data.iloc[train_idx]
    test_set = data.iloc[test_idx]

    return train_set, test_set


def split_gpm(gpm, fraction=1.0):
    """Split GenotypePhenotypeMap into two sets, a training and a test set.

    Parameters
    ----------
    data : pandas.DataFrame
        full dataset to split.

    fraction : float
        fraction in training set.

    Returns
    -------
    train_gpm : GenotypePhenotypeMap
        training set.

    test_gpm : GenotypePhenotypeMap
        test set.
    """
    train, test = split_data(gpm.data, fraction=fraction)

    train_gpm = GenotypePhenotypeMap.read_dataframe(
        train,
        wildtype=gpm.wildtype,
        mutations=gpm.mutations
    )

    test_gpm = GenotypePhenotypeMap.read_dataframe(
        test,
        wildtype=gpm.wildtype,
        mutations=gpm.mutations
    )

    return train_gpm, test_gpm



def gmean(x):
    """Calculate a geometric mean with zero and negative values.

    Following the gmean calculation from this paper:

    Habib, Elsayed AE. "Geometric mean for negative and zero values."
    International Journal of Research and Reviews in Applied Sciences 11
    (2012): 419-432.
    """
    x_neg = x[x < 0]
    x_pos = x[x > 0]
    x_zero = x[x == 0]

    n_neg = len(x_neg)
    n_pos = len(x_pos)
    n_zero = len(x_zero)
    N = len(x)

    gm_neg, gm_pos, gm_zero = 0, 0, 0
    if n_neg > 0:
        gm_neg = scipy.stats.mstats.gmean(abs(x_neg))
    if n_pos > 0:
        gm_pos = scipy.stats.mstats.gmean(x_pos)

    g1 = -1 * gm_neg * n_neg / N
    g2 = gm_pos * n_pos / N
    g3 = gm_zero * n_zero / N

    GM = g1 + g2 + g3
    return GM


def incremental_mean(old_mean, samples, M, N):
    """Calculate an incremental running mean.

    Parameters
    ----------
    old_mean : float or array
        current running mean(s) before adding samples
    samples : ndarray
        array containing the samples. Each column is a sample. Rows are
        independent values. Mean is taken across row.
    M : int
        number of samples in new chunk
    N : int
        number of previous samples in old mean
    """
    return ((N - M) * old_mean + samples.sum(axis=0)) / N


def incremental_var(old_mean, old_var, new_mean, samples, M, N):
    """Calculate an incremental variance.

    Parameters
    ----------
    old_mean : float or array
        current running mean(s) before adding samples
    old_var : float or array
        current running variance(s) before adding samples
    new_mean : float
        updated mean
    samples : ndarray
        array containing the samples. Each column is a sample. Rows are
        independent values. Mean is taken across row.
    M : int
        number of samples in new chunk
    N : int
        number of previous samples in old mean
    """
    return ((N - M) * old_var + np.array((samples - old_var) *
                                         (samples - new_mean)).sum(axis=0)) / N


def incremental_std(old_mean, old_std, new_mean, samples, M, N):
    """Calculate an incremental standard deviation.

    Parameters
    ----------
    old_mean : float or array
        current running mean(s) before adding samples
    samples : ndarray
        array containing the samples. Each column is a sample. Rows are
        independent values. Mean is taken across row.
    M : int
        number of samples in new chunk
    N : int
        number of previous samples in old mean
    """
    old_var = old_std**2
    return np.sqrt(incremental_var(old_mean, old_var, new_mean, samples, M, N))


def pearson(y_obs, y_pred):
    """ Calculate pearson coefficient between two variables.
    """
    x = y_obs
    y = y_pred

    xbar = np.mean(y_obs)
    ybar = np.mean(y_pred)

    terms = (x - xbar) * (y - ybar)

    numerator = sum(terms)

    # calculate denominator
    xdenom = sum((x - xbar)**2)
    ydenom = sum((y - ybar)**2)
    denominator = np.sqrt(xdenom) * np.sqrt(ydenom)

    return numerator / denominator


def rmsd(yobs, ypred):
    """Calculate the root mean squared deviation of an estimator."""
    ypred = np.array(ypred)
    yobs = np.array(yobs)
    return np.sqrt(np.sum((ypred - yobs)**2) / len(ypred))


def generalized_r2(y_obs, y_pred):
    """ Calculate the rquared between the observed and predicted y.
    See wikipedia definition of `coefficient of determination`.
    """
    # Mean fo the y observed
    y_obs_mean = np.mean(y_obs)
    # Total sum of the squares
    ss_total = sum((y_obs - y_obs_mean)**2)
    # Sum of squares of residuals
    ss_residuals = sum((y_obs - y_pred)**2)
    r_squared = 1 - (ss_residuals / ss_total)
    return r_squared


def explained_variance(y_obs, y_pred):
    """Returns the explained variance
    """
    # Mean fo the y observed
    y_obs_mean = np.mean(y_obs)
    # Total sum of the squares
    ss_total = sum((y_obs - y_obs_mean)**2)
    # Explained sum of squares
    ss_regression = sum((y_pred - y_obs_mean)**2)
    r_squared = (ss_regression / ss_total)
    return r_squared


def ss_residuals(y_obs, y_pred):
    """ calculate residuals """
    return sum((y_obs - y_pred)**2)


def chi_squared(y_obs, y_pred):
    """ Calculate the chi squared between observed and predicted y. """
    return sum((y_obs - y_pred)**2 / y_pred)

def aic(model):
    """Given a model, calculates an AIC score."""
    k = model.num_of_params
    L = model.lnlikelihood()
    return 2*(k-L)

# -----------------------------------------------------------------------
# Model error statistics
# -----------------------------------------------------------------------


def false_positive_rate(y_obs, y_pred, upper_ci, lower_ci, sigmas=2):
    """ Calculate the false positive rate of predicted values. Finds all values
    that equal zero in the known array and calculates the number of false
    positives found in the predicted given the number of samples and sigmas.

    The defined bounds are:
        (number of sigmas) * errors / sqrt(number of samples)

    Parameters
    ----------
    known : array-like
        Known values for comparing false positives
    predicted : array-like
        Predicted values
    errors : array-like
        Standard error from model
    n_samples : int
        number of replicate samples
    sigma : int (default=2)
        How many standard errors away (2 == 0.05 false positive rate)

    Returns
    -------
    rate : float
        False positive rate in data
    """

    N = len(y_obs)
    # Check that known, predicted, and errors are the same size.
    if N != len(y_pred) or N != len(upper_ci):
        raise Exception("Input arrays must all be the same size")

    # Number of known-zeros:
    known_zeros = 0

    # Number of false positives:
    false_positives = 0

    # Scale confidence bounds to the number of samples and sigmas
    upper_bounds = sigmas * upper_ci
    lower_bounds = sigmas * lower_ci

    for i in range(N):
        # Check that known value is zero
        if y_obs[i] == 0.0:

            # Add count to known_zero
            known_zeros += 1

            # Calculate bounds with given number of sigmas.
            upper = y_pred[i] + upper_bounds[i]
            lower = y_pred[i] - lower_bounds[i]

            # Check false positive rate.
            if y_obs[i] > upper or y_obs[i] < lower:
                false_positives += 1

    # Calculate false positive rate
    rate = false_positives / float(known_zeros)

    return rate


def false_negative_rate(y_obs, y_pred, upper_ci, lower_ci, sigmas=2):
    """ Calculate the false negative rate of predicted values. Finds all values
    that equal zero in the known array and calculates the number of false
    negatives found in the predicted given the number of samples and sigmas.

    The defined bounds are:
        (number of sigmas) * errors / sqrt(number of samples)

    Parameters
    ----------
    known : array-like
        Known values for comparing false negatives
    predicted : array-like
        Predicted values
    errors : array-like
        Standard error from model
    n_samples : int
        number of replicate samples
    sigma : int (default=2)
        How many standard errors away (2 == 0.05 false negative rate)

    Returns
    -------
    rate : float
        False negative rate in data
    """

    N = len(y_obs)
    # Check that known, predicted, and errors are the same size.
    if N != len(y_pred) or N != len(upper_ci):
        raise Exception("Input arrays must all be the same size")

    # Number of known-zeros:
    known_nonzeros = 0

    # Number of false negatives:
    false_negatives = 0

    # Scale confidence bounds to the number of samples and sigmas
    upper_bounds = sigmas * upper_ci
    lower_bounds = sigmas * lower_ci

    for i in range(N):
        # Check that known value is zero
        if y_obs[i] != 0.0:

            # Add count to known_zero
            known_nonzeros += 1

            # Calculate bounds with given number of sigmas.
            upper = y_pred[i] + upper_bounds[i]
            lower = y_pred[i] - lower_bounds[i]

            # Check false negative rate.
            if lower < 0 < upper:
                false_negatives += 1

    # Calculate false positive rate
    rate = false_negatives / float(known_nonzeros)

    return rate



from ..base import *
import numpy as np
import pytest

class TestBaseSimulation(object):
    
    wildtype = "00"
    mutations = {0:["0","1"], 1:["0","1"]}

    def test_init(self):
        sim = BaseSimulation(self.wildtype, self.mutations)
        # Checks
        assert isinstance(sim, BaseSimulation)
        
    def test_set_coefs_order(self):
        sim = BaseSimulation(self.wildtype, self.mutations)
        sim.set_coefs_order(2)
        # Tests 1: Check that epistasis map is attached
        assert hasattr(sim, "epistasis")
        # Test 2: Check epistasis is built
        assert hasattr(sim.epistasis, "sites")
        # Test 3: Sites have the correct length
        assert len(sim.epistasis.sites) == 4

    def test_set_coefs_sites(self):
        sim = BaseSimulation(self.wildtype, self.mutations)
        sites = [[0], [1], [2]]
        sim.set_coefs_sites(sites)
        # Tests 1: Check that epistasis map is attached
        assert hasattr(sim, "epistasis")
        # Test 2: Check epistasis is built
        assert hasattr(sim.epistasis, "sites")
        # Test 3: Sites have the correct length
        assert len(sim.epistasis.sites) == 3

    def test_build(self):
        sim = BaseSimulation(self.wildtype, self.mutations)
        with pytest.raises(Exception):
                sim.build()
